
ledger.api ?= {}

class AuthenticatedHttpClient extends @HttpClient

  constructor: (baseUrl) ->
    super(baseUrl)
    @_client = new HttpClient(baseUrl)

  setHttpHeader: (key, value) ->
    @_client.setHttpHeader(key, value)
    @

  jqAjax: (request) ->
    # We must use a jQuery Deferred object else the promise hierarchy becomes messy
    deferred = $.Deferred()

    # Check if we are authenticated
    if @isAuthenticated()
      # Even with an auth token the authenticated call can fail. Catch and authenticate if needed
      @_performUnsafeCall(request).then(_.bind(@_reportSuccess, @, request, deferred)).fail(_.bind(@_recoverUnsafeCallFailure, @, request, deferred))
    else
      # If not authenticated, then grab the authentication promise and perform the call after
      @_authenticateAndPerfomSafeCall(request, deferred)
    deferred

  # Performs an authentication and then a safe call (i.e. a call that cannot fail due to authentication issue)
  _authenticateAndPerfomSafeCall: (request, deferred) ->
    @_authenticate().then(_.bind(@_performSafeCall, @, request, deferred)).fail(_.bind(@_reportFailure, @, request, deferred))

  # Tries to recover from an usafe call (i.e. a call performed without being sure that the current auth token is valid)
  _recoverUnsafeCallFailure: (request, deferred, error) ->
    [jqXHR, textStatus, errorThrown] = error
    if jqXHR.status is 403 or jqXHR.status is 401
      @_authenticateAndPerfomSafeCall(request, deferred)
    else
      @_reportFailure(request, deferred, error)

  # Performs a safe call (i.e. a call performed rigth after an authentication)
  _performSafeCall: (request, deferred) ->
    @_client.setHttpHeader('X-LedgerWallet-AuthToken', @_authToken) if @_authToken?
    r = _(request).omit('success', 'error', 'complete')
    @_client
    .jqAjax(r)
    .done (data, textStatus, jqXHR) => @_reportSuccess(request, deferred, [data, textStatus, jqXHR])
    .fail (jqXHR, textStatus, errorThrown) => @_reportFailure(request, deferred, [jqXHR, textStatus, errorThrown])


  _performUnsafeCall: (request) ->
    @_client.setHttpHeader('X-LedgerWallet-AuthToken', @_authToken) if @_authToken?
    deferred = jQuery.Deferred()
    unsafeRequest = _(request).omit('success', 'error', 'complete')
    @_client
      .jqAjax(unsafeRequest)
      .done (data, textStatus, jqXHR) -> deferred.resolve([data, textStatus, jqXHR])
      .fail (jqXHR, textStatus, errorThrown) -> deferred.reject([jqXHR, textStatus, errorThrown])
    deferred

  # Properly reports HTTP failure
  _reportFailure: (request, deferred, error) ->
    [jqXHR, textStatus, errorThrown] = error
    deferred.reject(jqXHR, textStatus, errorThrown)
    request?.error?(jqXHR, textStatus, errorThrown)
    request?.complete?(jqXHR, textStatus)

  _reportSuccess: (request, deferred, success) ->
    [data, statusText, jqXHR] = success
    deferred.resolve(data, statusText, jqXHR)
    request?.success?(data, statusText, jqXHR)
    request?.complete?(jqXHR, statusText)

  isAuthenticated: -> if @_authToken? then yes else no

  _authenticate: () ->
    return @authenticationPromise if @authenticationPromise? and !@authenticationPromise.isFulfilled()
    d = ledger.defer()
    @authenticationPromise = d.promise
    @_performAuthenticate(d)
    @authenticationPromise

  _performAuthenticate: (deferred) ->
    bitidAddress = null
    deferred.retryNumber ?= 3
    ledger.bitcoin.bitid.getAddress()
    .fail (error) => throw [null, "Unable to get bitId address", error]
    .then (address) =>
      bitidAddress = address.bitcoinAddress.value
      @_client.jqAjax(type: "GET", url: "bitid/authenticate/#{bitidAddress}", dataType: 'json')
    .then (data) => ledger.bitcoin.bitid.signMessage(data['message'])
    .fail (error) =>
      throw error if _.isArray(error) and error.length is 3
      throw [null, "Unable to sign message", error]
    .then (signature) => @_client.jqAjax(type: "POST", url: 'bitid/authenticate', data: {address: bitidAddress, coin: ledger.config.network.ticker, signature: signature}, contentType: 'application/json', dataType: 'json')
    .then (data) =>
      @_authToken = data['token']
      ledger.app.emit 'wallet:authenticated'
      deferred.resolve()
    .fail (error) =>
      e "AUTH FAIL", error
      if deferred.retryNumber-- > 0 then @_performAuthenticate(deferred) else deferred.reject(error)
    .done()


  getAuthToken: (callback = null) ->
    d = ledger.defer(callback)
    if @isAuthenticated()
      d.resolve(@_authToken)
    else
      @_authenticate().then(-> d.resolve(@_authToken)).fail((ex) -> d.reject(ex))
    d.promise

  getAuthTokenSync: -> @_authToken


  @instance: (baseUrl = ledger.config.restClient.baseUrl) -> @_instance ?= new @(baseUrl)
  @reset: () -> @_instance = null

_.extend ledger.api,

  authenticated: (baseUrl = ledger.config.restClient.baseUrl) -> AuthenticatedHttpClient.instance(baseUrl)
  resetAuthentication: -> AuthenticatedHttpClient.reset()
